class Out:
    def __init__(self, id=None, name=None):
        super().__init__()
        self.id = id
        self.name = name

    def __eq__(self, other):
        return isinstance(other, type(self)) and (self.id, self.name) == (other.id, other.name)

    def __hash__(self):
        return hash((self.id, self.name))

    def __repr__(self):
        id = self.id if self.id is not None else ''
        name = self.name or ''
        return f'{name}${id}'


class OutDim:
    def __init__(self, out, dim):
        super().__init__()
        self.out = out
        self.dim = dim


class NodeArg:
    def __init__(self, name, value):
        super().__init__()
        self.name = name
        self.value = value

    def __repr__(self):
        return f'{self.name}={self.value}'


class NodeCall:
    def __init__(self, name, inputs, args):
        super().__init__()
        self.name = name
        self.inputs = inputs
        self.args = args

    def __repr__(self):
        if isinstance(self.inputs, list):
            inputs = '[' + ', '.join(str(i) for i in self.inputs) + ']'
        elif isinstance(self.inputs, dict):
            if len(self.inputs) == 1 and 'x' in self.inputs.keys():
                inputs = ''
            else:
                inputs = '{' + ', '.join(f'{k}={str(i)}'
                                         for k, i in self.inputs.items()) + '}'
        else:
            inputs = ''

        if self.args:
            args = '(' + ', '.join(str(a) for a in self.args.values()) + ')'
        else:
            args = ''

        return self.name + inputs + args


class GraphNode:
    def __init__(self, id, calls, shapes):
        super().__init__()
        self.id = id
        self.calls = calls
        self.shapes = shapes

        self._depth = 1

    def increse_depth(self):
        self._depth += 1

    def __repr__(self):
        indent = ' ' * (4 * self._depth)
        calls = ' + '.join(str(c) for c in self.calls)

        if not self.shapes:
            shapes = ''
        elif len(self.shapes) == 1 and 'x' in self.shapes.keys():
            shapes = str(self.shapes['x'])
        else:
            indent_shapes = indent + ' ' * 4
            shapes = ':\n' + '\n'.join(indent_shapes + k + str(s)
                                       for k, s in self.shapes.items())

        return f'{indent}{self.id}. {calls}{shapes}'



from enum import Enum, auto


class WithSet:
    def __init__(self, name, value):
        super().__init__()
        self.name = name
        self.value = value

    def __repr__(self):
        return f'set {self.name} = {self.value}'


class With:
    def __init__(self, name, variables):
        super().__init__()
        self.name = name
        self.variables = variables

        self._depth = 1

    def __repr__(self):
        indent = ' ' * (4 * self._depth)
        name = indent + f'with {self.name}:\n'

        indent += ' ' * 4
        variables = '\n'.join(indent + str(s) for s in self.variables.values())

        return name + variables


class NodeLetType(Enum):
    BOOL = auto()
    INT = auto()
    REAL = auto()
    DATA = auto()
    NODE = auto()
    DIM = auto()

    def __repr__(self):
        if self == self.BOOL:
            return 'bool'
        if self == self.INT:
            return 'int'
        if self == self.REAL:
            return 'real'
        if self == self.DATA:
            return 'data'
        if self == self.NODE:
            return 'node'
        if self == self.DIM:
            return 'dim'


TypeConverter = {
    NodeLetType.BOOL: lambda x: bool(x),
    NodeLetType.INT: lambda x: int(x),
    NodeLetType.REAL: lambda x: float(x),
}


class NodeLet:
    def __init__(self, name, shortcut, ty, value):
        super().__init__()
        self.name = name
        self.shortcut = shortcut
        self.ty = ty
        self.value = value

    def __repr__(self):
        shortcut = f'{self.shortcut}: ' if self.shortcut else ''
        value = self.value or '*'
        if self.ty == NodeLetType.DIM:
            ty_value = repr(self.ty)
        else:
            ty_value = f'{repr(self.ty)} {value}'
        return f'let {shortcut}{self.name} = {ty_value}'


class NodeType(Enum):
    DEFAULT = auto()
    EXTERN = auto()
    DATA = auto()
    EXEC = auto()

    def __repr__(self):
        if self == self.DEFAULT:
            return ''
        if self == self.EXTERN:
            return 'extern '
        if self == self.DATA:
            return 'data '
        if self == self.EXEC:
            return 'exec '


class Node:
    def __init__(self, name, variables, withs, children, graph):
        super().__init__()
        self.name = name
        self.variables = variables
        self.withs = withs
        self.children = children
        self.graph = graph

        self.ty = NodeType.DEFAULT

        self._depth = 0

        for c in self.children.values():
            c.increse_depth()

    def increse_depth(self):
        for c in self.children.values():
            c.increse_depth()
        for g in self.graph.values():
            g.increse_depth()

        self._depth += 1

    def __repr__(self):
        indent = ' ' * (4 * self._depth)
        name = indent + f'{repr(self.ty)}node {self.name}:\n'

        indent += ' ' * 4
        variables = '\n'.join(indent + str(v)
                              for v in self.variables.values()) + '\n'
        withs = '\n'.join(str(w) for w in self.withs.values()) + '\n'
        children = '\n'.join(str(c) for c in self.children.values()) + '\n'
        graph = '\n'.join(str(c) for c in self.graph.values()) + '\n'

        return name + variables + withs + children + graph



import torch.nn as nn

from n3 import ExternNode


class Relu(ExternNode):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._inner = nn.ReLU(inplace=True)

    def forward(self, x):
        return self._inner(x)



import os
from pathlib import Path

import click

from ..builder import NodeRoot
from ..std import trim_name

DATA_DIR = 'data'
MODELS_DIR = 'models'
NODES_DIR = 'nodes'
NODES_USER_DIR = '__user__'


class ExecRoot:
    def __init__(self, env):
        super().__init__()
        self.env = env
        self.inner = NodeRoot()

        self._create_root_dir()
        self._load_local_nodes()

    @property
    def root_dir(self):
        return self.env.root

    def _create_root_dir(self):
        path = Path(self.root_dir)

        if path.exists():
            if path.is_dir():
                return
            raise Exception(f'not directory: {path}')

        if not click.confirm(
            f'''It seems that there is no root directory on "{path}"
- Do you want to create one?''', default=False):
            raise Exception('user cancelled')

        path.mkdir(parents=True, exist_ok=True)
        for name in [DATA_DIR, MODELS_DIR, NODES_DIR, f'{NODES_DIR}/{NODES_USER_DIR}']:
            (path / name).mkdir(exist_ok=True)

    def _load_local_nodes(self):
        path = os.path.join(self.root_dir, 'nodes')
        for p in Path(path).glob('*.n3'):
            name = trim_name(p.name)
            self.inner.add_source_path(name, str(p))



import torch.nn as nn

from .. import ast


class ExternNode(nn.Module):
    def __init__(self, input, output, **kwargs):
        super().__init__()
        self._node_input = input
        self._node_output = output

        for k, v in kwargs.items():
            setattr(self, k, v)

    def forward(self, **kwargs):
        name = self.__class__.__name__
        raise NotImplementedError(f'extern node {name} is not implemented yet')

    def __repr__(self, depth=0):
        indent = ' ' * (depth * 4) + ' ' * 2
        indent_node = ' ' * ((depth+1) * 4)

        name = f'[node extern object {self.__class__.__name__}]'
        input = f'\n{indent}[input]\n' + \
            '\n'.join(f'{indent_node}{k}: {repr(v)}'
                      for k, v in self._node_input.items())
        output = f'\n{indent}[output]\n' + \
            '\n'.join(f'{indent_node}{k}: {repr(v)}'
                      for k, v in self._node_output.items())
        return name + input + output


class NodeExecutable(nn.Module):
    def __init__(self, name, input, output, tensor_graph):
        super().__init__()
        self._name = name
        self._node_input = input
        self._node_output = output
        self._tensor_graph = tensor_graph

    def __call__(self, **kwargs):
        output = {ast.Out(0, k): x for k, x in kwargs.items()}

        for node in self._tensor_graph:
            x = {k: output[n] for k, n in node._node_input.items()}
            x = node(**x)
            if not isinstance(x, dict):
                x = {'x': x}
            x = {n: x[k] for k, n in node._node_output.items()}
            output = {**output, **x}

        return {k.name: v for k, v in x.items()}

    def __repr__(self, depth=0):
        indent = ' ' * (depth * 4) + ' ' * 2
        indent_node = ' ' * ((depth+1) * 4)

        prefix = '' if depth else '* '

        name = f'{prefix}[node object {self._name}]'
        input = f'\n{indent}[input]\n' + \
            '\n'.join(f'{indent_node}{k}: {repr(v)}'
                      for k, v in self._node_input.items())
        output = f'\n{indent}[output]\n' + \
            '\n'.join(f'{indent_node}{k}: {repr(v)}'
                      for k, v in self._node_output.items())
        tensor_graph = '\n'.join(f'{indent}({id}) {n.__repr__(depth+1)}'
                                 for id, n in enumerate(self._tensor_graph))
        return name + input + output + '\n' + tensor_graph



from .. import ast


def replace(names, value, variables, shortcuts, is_root=False):
    if value is None:
        return None
    elif isinstance(value, ast.Variable):
        if value.name in names:
            names = ', '.join(reversed(names))
            raise Exception(f'cycled variable: {names}')
        if value.name in shortcuts.keys():
            value.name = shortcuts[value.name]
        if value.name in variables.keys():
            value = variables[value.name]

        names.append(value.name)
        value.value = replace(names, value.value, variables, shortcuts)
        names.pop()
        return value
    elif isinstance(value, ast.Expr):
        value.lhs = replace(names, value.lhs, variables, shortcuts)
        value.rhs = replace(names, value.rhs, variables, shortcuts)
    return value


def build(value):
    if isinstance(value, ast.Variable):
        if value.is_hint():
            return value
        if value.value is None:
            raise Exception(f'value is needed: {value.name}')
        return ast.TypeConverter[value.ty](build(value.value))
    if isinstance(value, ast.Expr):
        lhs = build(value.lhs)
        rhs = build(value.rhs)
        if ast.is_hint(lhs) or ast.is_hint(rhs):
            return ast.Expr(value.op, lhs, rhs)
        return ast.OperatorFn[value.op](lhs, rhs)
    if isinstance(value, list):
        return [build(v) for v in value]
    if isinstance(value, dict):
        return {k: build(v) for k, v in value.items()}
    if isinstance(value, ast.Shape):
        dims = [int(v) for v in build(value.dims)]
        return ast.Shape(dims, value.name)
    return value


class Graph:
    def __init__(self):
        super().__init__()
        self._shortcuts = {}
        self._variables = {}

    def add(self, variable):
        if variable.name in self._variables.keys():
            raise Exception(f'duplicated variable: {variable.name}')
        self._variables[variable.name] = variable

    def get(self, name):
        return self._variables[name]

    def build(self):
        shortcuts_map = {v.shortcut: name
                         for name, v in self._variables.items()
                         if v.shortcut}
        self._variables = {name: replace([], v, self._variables, shortcuts_map, is_root=True)
                           for name, v in self._variables.items()}
        self._shortcuts = {v.shortcut or v.name: v
                           for v in self._variables.values()}

    def hint(self, out, dims):
        def _hint(out, dim, value, is_root=False):
            if isinstance(value, ast.Variable):
                if value.name not in self._shortcuts.keys():
                    candidates = ', '.join(self._shortcuts.keys())
                    raise Exception(
                        f'no such variable: {value.name}, candidates are: {candidates}')
                output = self._shortcuts[value.name]
                # hint in-place
                if output.ty == ast.NodeLetType.DIM and is_root:
                    output.value = ast.OutDim(out, dim)
                return output
            if isinstance(value, ast.Expr):
                value.lhs = _hint(out, dim, value.lhs)
                value.rhs = _hint(out, dim, value.rhs)
            return value

        return [_hint(out, dim, v, is_root=True) for dim, v in enumerate(dims)]

    def replace(self, value):
        if value is None:
            return None
        elif isinstance(value, ast.Variable):
            if value.name in self._shortcuts.keys():
                return self._shortcuts[value.name]
            raise Exception(f'no such variable: {value.name}')
        elif isinstance(value, ast.Expr):
            value.lhs = self.replace(value.lhs)
            value.rhs = self.replace(value.rhs)
        return value

    def apply(self, variables, is_shortcut=False):
        # TODO: test types
        self_variables = self._shortcuts if is_shortcut else self._variables
        for k, v in variables.items():
            self_variables[k].value = v

    def assert_estimable(self):
        all(v.assert_estimable() for v in self._variables.values())

    def __repr__(self):
        return '\n'.join(f'let {v}' for v in self._variables.values())



from abc import *


class Use(metaclass=ABCMeta):
    def __init__(self, name):
        super().__init__()
        self.name = name

    def __repr__(self):
        return f'use {self.name}'


class UseByLocal(Use):
    def __init__(self, name, *_):
        super().__init__(name)


class UseByAuthor(Use):
    def __init__(self, name, author):
        super().__init__(name)
        self.author = author

    def __repr__(self):
        return f'{super().__repr__()} by {self.author}'


class UseByWeb(Use):
    def __init__(self, name, url):
        super().__init__(name)
        self.url = url

    def __repr__(self):
        return f'{super().__repr__()} by w{self.url}'



from .file import File
from .graph import GraphNode, NodeArg, NodeCall, Out, OutDim
from .node import Node, NodeLet, NodeLetType, NodeType, TypeConverter, With, WithSet
from .use import Use, UseByLocal, UseByAuthor, UseByWeb
from .variable import Expr, Operator, OperatorFn, Shape, Variable, is_estimable, is_hint



import inflection

from .variable import build


class TensorGraph:
    def __init__(self, id, name, kwargs, input, output, input_shapes, output_shapes):
        super().__init__()
        self.id = id
        self.name = name
        self.kwargs = kwargs or {}
        self.input = input or {}
        self.output = output or {}

        self.input_shapes = input_shapes or {
            k: None for k in self.input.keys()}
        self.output_shapes = output_shapes or {
            k: None for k in self.output.keys()}

    def get_input_shapes(self):
        return self.input_shapes

    def get_output_shapes(self):
        return self.output_shapes

    def build(self, root):
        kwargs = {inflection.underscore(k.title().replace(' ', '_')): build(v)
                  for k, v in self.kwargs.items()}

        node_cls = root.get_extern(self.name)
        return node_cls(input=self.input, output=self.output, **kwargs)

    def __repr__(self):
        kwargs = '\n\t[args]\n' + \
            '\n'.join(f'\t\t{repr(v)}' for v in self.kwargs.values())
        input = '\n\t[input]\n' + \
            '\n'.join(f'\t\t{k}: {repr(v)}{self.input_shapes[k]}'
                      for k, v in self.input.items())
        output = '\n\t[output]\n' + \
            '\n'.join(f'\t\t{k}: {repr(v)}{self.output_shapes[k]}'
                      for k, v in self.output.items())
        return self.name + kwargs + input + output



from copy import deepcopy
import importlib.util

import torch.nn as nn

from .. import ast, std
from ..parser import Parser
from .extern_node import ExternNode, NodeExecutable
from .tensor import TensorGraph
from .variable import Graph
from .variable import build as build_variable


_BuiltinsInput = ['AssertShape', ]
_BuiltinsNode = ['Transform', ]


def make_graph(variables):
    graph = Graph()
    for let in variables.values():
        var = ast.Variable(let.name, let.value)
        var.shortcut = let.shortcut
        var.ty = let.ty
        graph.add(var)
    graph.build()
    return graph


def _assert_tensor_graph_name(node: ast.GraphNode, expects, sized=False):
    expects = ' | '.join(expects)

    num_calls = len(node.calls)
    assert num_calls == 1, f'Expected {expects} node, Given {num_calls} nodes'
    call = node.calls[0]
    assert call.name in expects, f'Expected {expects} node, Given {call.name}'
    assert call.inputs is None, f'Expected {expects} node, Given {repr(call)}'
    assert call.args is None, f'Expected {expects} node, Given {repr(call)}'
    if sized:
        assert node.shapes is not None, f'Expected sized node: {call.name}'
    return call


def _merge_shapes(last_outputs, new_inputs):
    for name in last_outputs.keys():
        last_output = last_outputs[name]
        new_input = new_inputs[name]

        # dynamic size
        if not new_input:
            new_inputs[name] = last_output
            continue

        assert len(last_output) == len(new_input), \
            f'Expected {len(new_input)} dims, Given {len(last_output)} dims'

        for dim in range(len(new_input)):
            last_dim = last_output.dims[dim]
            new_dim = new_input.dims[dim]
            if not ast.is_hint(last_dim):
                # replace
                if ast.is_hint(new_dim):
                    new_dim.ty = ast.NodeLetType.INT
                    new_dim.value = last_dim
                # test value
                elif ast.is_estimable(last_dim) and ast.is_estimable(new_dim):
                    last_dim = build_variable(last_dim)
                    new_dim = build_variable(new_dim)
                    assert last_dim == new_dim, f'Expected {new_dim}, Given {last_dim}'


class NodeIR:
    def __init__(self, name, graph, tensor_graph):
        super().__init__()
        self.name = name
        self.graph = graph
        self.tensor_graph = tensor_graph

        self.id = None
        self.input = None
        self.output = None

    def get_input_shapes(self):
        input_node = self.tensor_graph[0]
        if input_node.name in _BuiltinsInput:
            return input_node.get_output_shapes()
        return input_node.get_input_shapes()

    def get_output_shapes(self):
        for output_node in reversed(self.tensor_graph):
            shapes = output_node.get_output_shapes()
            if shapes:
                # filter dynamic size
                if len(shapes) == 1 and 'x' in shapes.keys():
                    if not shapes['x']:
                        continue
                return shapes
        return self.tensor_graph[-1].get_output_shapes()

    def apply_variables(self, variables, is_shortcut=False):
        self.graph.apply(variables, is_shortcut=is_shortcut)

    def build(self, root):
        input = self.input or {k: ast.Out(0, k)
                               for k in self.get_input_shapes().keys()}
        output = self.output or {k: ast.Out(1, k)
                                 for k in self.get_output_shapes().keys()}
        tensor_graph = [n.build(root) for n in self.tensor_graph]
        return NodeExecutable(self.name, input, output, tensor_graph)

    def __repr__(self):
        name = f'[node ir {self.name}]'
        graph = repr(self.graph)
        tensor_graph = '\n'.join(f'({id}) {repr(n)}'
                                 for id, n in enumerate(self.tensor_graph))
        return f'{name}\n{graph}\n\n{tensor_graph}'


class NodeContainer:
    def __init__(self, parent, name):
        super().__init__()
        self.parent = parent
        self.name = name

        self.graph = None
        self.tensor_graph = []

        self.children = {}

        self.last_id = 0

    def add_use(self, use: ast.Use):
        # Step 1. get the source
        # Step 2. build
        # Step 3. store
        raise NotImplementedError()

    def add_with(self, w: ast.With):
        # Step 1. get the node
        node = self.get(w.name)

        # Step 2. apply variables
        args = {k: self.graph.replace(v.value)
                for k, v in w.variables.items()}
        node.apply_variables(args)

        # Step 3. store
        self.children[w.name] = node

    def add_child(self, child: ast.Node):
        # Step 1. convert to file
        file = ast.File([], child)

        # Step 2. build
        node = self._build(file, self)

        # Step 3. store
        self.children[child.name] = node

    def add_tensor_graph(self, node: ast.GraphNode):
        # input node
        if node.id == 0:
            call = _assert_tensor_graph_name(node, ['Input'], sized=True)

            output = {s: ast.Out(0, s) for s in node.shapes.keys()}

            callee = TensorGraph(
                node.id, 'AssertShape', None, output, output, None, node.shapes)
            self.tensor_graph.append(callee)
        else:
            # test builtins
            call = node.calls[0]
            if call.name in _BuiltinsNode:
                callee = self._build_builtins(call.name, node)
                self.tensor_graph.append(callee)
                self.last_id = node.id
                return

            for call in node.calls:
                # Step 1. get the node
                callee = self.get(call.name)
                callee.id = node.id

                # Step 2. apply variables
                if call.args:
                    args = {k: self.graph.replace(v.value)
                            for k, v in call.args.items()}
                    callee.apply_variables(args, is_shortcut=True)

                # Step 3. apply IO
                inputs = call.inputs or {}
                for k in callee.get_input_shapes().keys():
                    if k not in inputs.keys():
                        inputs[k] = ast.Out(name=k)
                outputs = {k: ast.Out(node.id, k)
                           for k in callee.get_output_shapes().keys()}

                # set default IO to "x"
                inputs = inputs or {'x': ast.Out(name='x')}
                outputs = outputs or {'x': ast.Out(id=node.id, name='x')}

                callee.input = inputs
                callee.output = outputs

                # Step 4. merge shapes
                if self.tensor_graph:
                    last_outputs = {k: self._get_shape(x)
                                    for k, x in inputs.items()}
                    new_inputs = callee.get_input_shapes()

                    _merge_shapes(last_outputs, new_inputs)

                    # identity
                    new_outputs = callee.get_output_shapes()
                    for name in new_outputs.keys():
                        if not new_outputs[name]:
                            new_outputs[name] = new_inputs[name]
                elif inputs:
                    for x in inputs.values():
                        x.id = 0

                # Step 5. store
                self.tensor_graph.append(callee)

            # Step 6. merge dedicated shapes
            if node.shapes:
                outputs = self.get_output_shapes()
                _merge_shapes(node.shapes, outputs)

            # Step 7. store id
            self.last_id = node.id

    def get_output_shapes(self):
        for node in reversed(self.tensor_graph):
            outputs = node.get_output_shapes()
            if outputs:
                return outputs

    def _get_shape(self, out):
        for node in reversed(self.tensor_graph):
            # test id
            if out.id is not None:
                if node.id > out.id:
                    continue
                if node.id < out.id:
                    break

            shapes = node.get_output_shapes()
            if out.name in shapes.keys():
                out.id = node.id
                return shapes[out.name]
        raise Exception(f'no such input: {repr(out)}')

    def _update_out_id(self, out):
        if out.id is not None:
            return out
        for node in reversed(self.tensor_graph):
            shapes = node.get_output_shapes()
            if out.name in shapes.keys():
                out.id = node.id
                return out
        raise Exception(f'no such input: {repr(out)}')

    def get(self, name):
        if name in self.children.keys():
            return deepcopy(self.children[name])
        return self.parent.get(name)

    def build(self):
        return NodeIR(self.name, self.graph, self.tensor_graph)

    def _build(self, file, parent=None):
        return self.parent._build(file, parent)

    def _build_builtins(self, name, node):
        if name == 'Transform':
            return self._build_transform(node)
        raise NotImplementedError

    def _build_transform(self, node):
        call = _assert_tensor_graph_name(node, ['Transform'], sized=True)

        # Step 1. get the IO
        inputs = self.get_output_shapes()
        outputs = node.shapes

        # Step 2. match the tuple
        input_keys = inputs.keys()
        output_keys = outputs.keys()
        if input_keys != output_keys:
            input_keys = ', '.join(input_keys)
            output_keys = ', '.join(output_keys)
            raise Exception(
                f'Expected tuple ({input_keys}), Given ({output_keys})')

        # Step 3. match the size
        for name in input_keys:
            input = build_variable(inputs[name].product())
            output = build_variable(outputs[name].product())
            if ast.is_hint(input) or ast.is_hint(output):
                # TODO: FUTURE: implement comparing hinted values
                print('warning: comparing hinted values is not supported yet!')
            else:
                assert input == output, f'Expected product {input}, Given {output}'

        # Step 4. apply to tensor graph
        id = self.last_id + 1
        input = {s: self._update_out_id(ast.Out(name=s))
                 for s in inputs.keys()}
        output = {s: ast.Out(id, s)
                  for s in outputs.keys()}

        # Step 5. store
        kwargs = {'output_shapes': outputs}
        return TensorGraph(node.id, call.name, kwargs, input, output, inputs, outputs)


class ExternNodeContainer:
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.graph = None

        self.input = None
        self.output = None

    def add_tensor_graph(self, node: ast.GraphNode):
        name = 'Input' if node.id == 0 else 'Output'
        call = _assert_tensor_graph_name(node, [name])

        if node.id == 0:
            self.input = node.shapes
        elif node.id == 1:
            self.output = node.shapes

    def build(self):
        # set default IO to "x"
        input = {s: ast.Out(0, s) for s in self.input.keys()} \
            if self.input else {'x': ast.Out(0, 'x')}
        output = {s: ast.Out(1, s) for s in self.output.keys()} \
            if self.output else {'x': ast.Out(1, 'x')}
        tensor_graph = TensorGraph(
            0, self.name, self.graph._variables, input, output, self.input, self.output)
        return NodeIR(self.name, self.graph, [tensor_graph])


class NodeRoot:
    def __init__(self):
        super().__init__()
        self._source_paths = {}
        self._sources = {}
        self._irs = {}

        self._externs = std.externs
        self._extern_nodes = {}

        self._parser = Parser()

        for source in std.sources.values():
            self.add_source(source)

    def add_source(self, source):
        file = self._parser.parse(source)
        self._sources[file.node.name] = file

    def add_source_path(self, name, path):
        self._source_paths[name] = path

    def add_extern(self, name, path):
        self._externs[name] = path

    def get(self, name):
        if name in self._irs.keys():
            return deepcopy(self._irs[name])

        if name in self._source_paths.keys():
            path = self._source_paths.pop(name)
            with open(path) as f:
                self.add_source(f.read())

        if name in self._sources.keys():
            result = self._irs[name] = self._build(self._sources.pop(name))
            return deepcopy(result)
        raise Exception(f'undefined node: {name}')

    def get_extern(self, name):
        if name in self._extern_nodes.keys():
            return self._extern_nodes[name]

        if name in self._externs.keys():
            path = self._externs.pop(name)

            loader = importlib.machinery.SourceFileLoader(f'ext_{name}', path)
            spec = importlib.util.spec_from_loader(loader.name, loader)
            mod = importlib.util.module_from_spec(spec)
            loader.exec_module(mod)

            node_cls = self._extern_nodes[name] = getattr(mod, name)
            return node_cls
        raise Exception(f'undefined extern node: {name}')

    def _build(self, file: ast.File, parent: NodeContainer = None):
        if file.node.ty == ast.NodeType.EXTERN:
            return self._build_extern(file)

        container = NodeContainer(parent or self, file.node.name)

        # Step 1. import remote models
        for use in file.uses:
            container.add_use(use)

        # Step 2. make a graph
        container.graph = make_graph(file.node.variables)

        # Step 3. hint variables with tensor graph
        for id, node in file.node.graph.items():
            if node.shapes:
                for x, shapes in node.shapes.items():
                    shapes.dims = container.graph.hint(
                        ast.Out(id, x), shapes.dims)

        # Step 4. re-define nodes (with)
        for w in file.node.withs.values():
            container.add_with(w)

        # Step 5. build children nodes
        for child in file.node.children.values():
            container.add_child(child)

        # Step 6. make a tensor graph
        last_id = 0
        for id, node in file.node.graph.items():
            if id - last_id != 1 and not (last_id == 0 and id == 0):
                raise Exception(f'Expected id {last_id+1}, Given {id}')
            container.add_tensor_graph(node)
            last_id = id

        # Step 7. store
        return container.build()

    def _build_extern(self, file: ast.File):
        container = ExternNodeContainer(file.node.name)

        # Step 1. make a graph
        container.graph = make_graph(file.node.variables)

        # Step 2. hint variables with tensor graph
        for id, node in file.node.graph.items():
            if node.shapes:
                for x, shapes in node.shapes.items():
                    shapes.dims = container.graph.hint(
                        ast.Out(id, x), shapes.dims)

        # Step 3. make a tensor graph
        num_tensor_graph = len(file.node.graph)
        assert num_tensor_graph == 2, f'Expected Input & Output node, Given {num_tensor_graph}'

        last_id = -1
        for id, node in file.node.graph.items():
            if id - last_id != 1:
                raise Exception(f'Expected id {last_id+1}, Given {id}')
            container.add_tensor_graph(node)
            last_id = id

        # Step 4. store
        return container.build()



import os
from pathlib import Path


def trim_name(name):
    return name.split('.')[0].replace('_', '-').title().replace('-', '')


def _read_file(f):
    with open(f, encoding='utf-8') as f:
        return f.read()


root_dir = os.path.dirname(__file__)

externs = {trim_name(f.name): str(f)
           for f in Path(root_dir).rglob(r'[A-Za-z]*.py')}
sources = {trim_name(f.name): _read_file(f)
           for f in Path(root_dir).rglob(r'*.n3')}



import torch.nn as nn

from n3 import ExternNode


class Conv2D(ExternNode):
    kernel_size: int
    padding: int

    stride: int

    input_channels: int
    output_channels: int

    bias: bool

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._inner = nn.Conv2d(self.input_channels,
                                self.output_channels,
                                kernel_size=self.kernel_size,
                                stride=self.stride,
                                padding=self.padding,
                                bias=self.bias)

    def forward(self, x):
        return self._inner(x)



# Many of this code is from:
# https://github.com/lark-parser/lark/blob/master/examples/python_parser.py

from lark import Lark, Transformer
from lark.indenter import Indenter

from .. import ast


class _Indenter(Indenter):
    NL_type = '_NEWLINE'
    OPEN_PAREN_types = ['LPAR', 'LSQB', 'LBRACE']
    CLOSE_PAREN_types = ['RPAR', 'RSQB', 'RBRACE']
    INDENT_type = '_INDENT'
    DEDENT_type = '_DEDENT'
    tab_len = 8


class _Transformer(Transformer):
    _USE_TYPES = {
        'use_by_local': ast.UseByLocal,
        'use_by_author': ast.UseByAuthor,
        'use_by_web': ast.UseByWeb,
    }

    _OPS = {
        '+': ast.Operator.ADD,
        '-': ast.Operator.SUB,
        '*': ast.Operator.MUL,
        '/': ast.Operator.DIV,
        '%': ast.Operator.MOD,
    }

    def file_input(self, args):
        return ast.File(args[:-1], args[-1])

    def rootdef(self, args):
        args[1].ty = args[0]
        return args[1]

    def node_type_extern(self, args):
        return ast.NodeType.EXTERN

    def node_type_data(self, args):
        return ast.NodeType.DATA

    def node_type_exec(self, args):
        return ast.NodeType.EXEC

    def node_type_default(self, args):
        return ast.NodeType.DEFAULT

    def use(self, args):
        name = str(args[0])
        by = args[1]
        by_method = by.data
        by_source = str(by.children[0]) if by.data != 'use_by_local' else None
        return self._USE_TYPES[by_method](name, by_source)

    def nodedef(self, args):
        return ast.Node(str(args[0]), *args[1:])

    def node_suite_let(self, args):
        # TODO: check duplicated names
        return {a.name: a for a in args}

    def node_suite_with(self, args):
        # TODO: check duplicated names
        return {a.name: a for a in args}

    def node_suite_children(self, args):
        # TODO: check duplicated names
        return {a.name: a for a in args}

    def node_suite_graph(self, args):
        # TODO: check duplicated names
        return {a.id: a for a in args}

    def nodelet(self, args):
        name = str(args[1])
        shortcut = str(args[0]) if args[0] is not None else None
        ty, value = args[2]
        return ast.NodeLet(name, shortcut, ty, value)

    def nodelet_type_bool(self, args):
        return ast.NodeLetType.BOOL, args[0] if args else None

    def nodelet_type_int(self, args):
        return ast.NodeLetType.INT, args[0] if args else None

    def nodelet_type_real(self, args):
        return ast.NodeLetType.REAL, args[0] if args else None

    def nodelet_type_data(self, args):
        return ast.NodeLetType.DATA, str(args[0]) if args else None

    def nodelet_type_node(self, args):
        return ast.NodeLetType.NODE, str(args[0]) if args else None

    def nodelet_type_dim(self, args):
        return ast.NodeLetType.DIM, None

    def graph(self, args):
        return ast.GraphNode(args[0], args[1], args[2] if len(args) > 2 else None)

    def graph_nodes(self, args):
        return args

    def graph_node(self, args):
        return ast.NodeCall(str(args[0]), args[1], args[2])

    def graph_inputs_dict(self, args):
        # TODO: check duplicated names
        return {a[0]: a[1] for a in args}

    def graph_inputs_list(self, args):
        return args

    def graph_input_kw(self, args):
        return str(args[0]), args[1]

    def graph_args(self, args):
        # TODO: check duplicated names
        return {a.name: a for a in args}

    def graph_arg(self, args):
        return ast.NodeArg(str(args[0]), args[1])

    def graph_shapes(self, args):
        # TODO: check duplicated names
        return {a.name: a for a in args}

    def graph_shape_kw(self, args):
        args[1].name = str(args[0])
        return args[1]

    def graph_shape(self, args):
        return ast.Shape(args)

    def dim(self, args):
        return args[0] if args else None

    def out(self, args):
        return ast.Out(args[1], str(args[0] or 'x'))

    def out_const(self, args):
        return args[0]

    def withdef(self, args):
        return ast.With(str(args[0]), args[1])

    def with_suite(self, args):
        # TODO: check duplicated names
        return {a.name: a for a in args}

    def withset(self, args):
        return ast.WithSet(args[0], args[1])

    def expr(self, args):
        return ast.Expr(ast.Operator.OR, args[0], args[1])

    def xor_expr(self, args):
        return ast.Expr(ast.Operator.XOR, args[0], args[1])

    def and_expr(self, args):
        return ast.Expr(ast.Operator.AND, args[0], args[1])

    def arith_expr(self, args):
        return ast.Expr(self._OPS[args[1]], args[0], args[2])

    def term(self, args):
        return ast.Expr(self._OPS[args[1]], args[0], args[2])

    def factor(self, args):
        return ast.Expr(ast.Operator.NEG, args[1])

    def power(self, args):
        return ast.Expr(ast.Operator.POW, args[0], args[1])

    def variable(self, args):
        return ast.Variable(str(args[0]))

    def fullname(self, args):
        return ' '.join(args)

    def DEC_NUMBER(self, value):
        return int(value)

    def FLOAT_NUMBER(self, value):
        return float(value)


class Parser:
    def __init__(self):
        super().__init__()
        self._inner = Lark.open(
            'grammar.lark',
            parser='lalr',
            rel_to=__file__,
            postlex=_Indenter(),
            start='file_input',
            maybe_placeholders=True,
            cache=True,
        )
        self._transformer = _Transformer()

    def parse(self, source: str):
        tree = self._inner.parse(source + '\n')
        return self._transformer.transform(tree)



import argparse
import sys
import types

from .eval import eval
from .train import train
from .variable import EnvVars


def route():
    # get the module itself, enviroment variables
    module = globals()
    env = EnvVars()

    # get the usable executions
    modes = [k for k, v in module.items()
             if isinstance(v, types.FunctionType) and v.__module__ == f'n3.exec.{k}']

    # define parser
    parser = argparse.ArgumentParser('n3',
                                     add_help=False,
                                     description='Process some integers.')
    parser.add_argument('help', type=bool, nargs='?',
                        help='show this help message and exit')
    parser.add_argument('mode', type=str, choices=modes,
                        help='hehehehe')  # TODO: add help message
    parser.add_argument('exec', type=str,
                        help='hehehehe')  # TODO: add help message

    # parser can modify envs
    env.attach_parser(parser)

    # show help message
    args = sys.argv[1:]
    if not args or len(args) == 1 and args[0] == 'help':
        parser.print_help()
        return

    # parse
    args, _ = parser.parse_known_args()
    env.apply(args)

    # execute
    module[args.mode](env, args.exec, help=args.help is True)



import torch.nn as nn

from n3 import ExternNode


class Dropout(ExternNode):
    probability: float

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._inner = nn.Dropout(self.probability, inplace=True)

    def forward(self, x):
        return self._inner(x)



class File:
    def __init__(self, uses, node):
        super().__init__()
        self.uses = uses
        self.node = node

    def __repr__(self):
        uses = '\n'.join(str(u) for u in self.uses)
        return uses + f'\n{self.node}'



from setuptools import find_packages, setup

author = {
    'name': 'Ho Kim',
    'email': 'ho.kim@gnu.ac.kr',
}


def read(filename):
    with open(filename, 'r', encoding='utf-8') as f:
        return f.read()


def read_requirements():
    rs = read('requirements.txt').split('\n')
    rs = [r.strip() for r in rs]
    rs = [r for r in rs if len(r) > 0 and not r.startswith('#')]
    return rs


setup(
    version=read('VERSION'),

    name='n3',
    description='TODO',  # TODO: to be implemented
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url=r'https://github.com/kerryeon/n3',

    author=author['name'],
    author_email=author['email'],
    maintainer=author['name'],
    maintainer_email=author['email'],
    license='MIT',

    install_requires=read_requirements(),
    packages=find_packages(),

    entry_points={
        'console_scripts': [
            'n3=n3.exec:route',
        ],
    },

    include_package_data=True,
    zip_safe=False,
)



from enum import Enum, auto

from .graph import Out
from .node import NodeLetType


def is_estimable(value):
    if isinstance(value, (Variable, Expr)):
        return value.is_estimable()
    return value is not None


def is_hint(value):
    if isinstance(value, (Variable, Expr)):
        return value.is_hint()
    return isinstance(value, Out)


class Variable:
    def __init__(self, name, value=None):
        super().__init__()
        self.name = name
        self.shortcut = None
        self.ty = None
        self.value = value

    def assert_estimable(self):
        assert is_estimable(self.value), f'unestimable variable: {self.name}'
        return True

    def is_estimable(self):
        return is_estimable(self.value)

    def is_hint(self):
        return self.ty == NodeLetType.DIM or is_hint(self.value)

    def __repr__(self):
        if self.value is not None:
            return f'{self.name}={repr(self.value)}'
        return self.name


class Operator(Enum):
    NEG = auto()
    ADD = auto()
    SUB = auto()
    MUL = auto()
    DIV = auto()
    MOD = auto()
    POW = auto()
    AND = auto()
    OR = auto()
    XOR = auto()

    def __repr__(self):
        if self == self.NEG:
            return '-'
        if self == self.ADD:
            return '+'
        if self == self.SUB:
            return '-'
        if self == self.MUL:
            return '*'
        if self == self.DIV:
            return '/'
        if self == self.MOD:
            return '%'
        if self == self.POW:
            return '**'
        if self == self.AND:
            return '&'
        if self == self.OR:
            return '|'
        if self == self.XOR:
            return '^'


OperatorFn = {
    Operator.NEG: lambda a, _: -a,
    Operator.ADD: lambda a, b: a + b,
    Operator.SUB: lambda a, b: a - b,
    Operator.MUL: lambda a, b: a * b,
    Operator.DIV: lambda a, b: a / b,
    Operator.MOD: lambda a, b: a % b,
    Operator.POW: lambda a, b: a ** b,
    Operator.AND: lambda a, b: a & b,
    Operator.OR: lambda a, b: a ^ b,
    Operator.XOR: lambda a, b: a ^ b,
}


class Expr:
    def __init__(self, op, lhs, rhs=None):
        super().__init__()
        self.op = op
        self.lhs = lhs
        self.rhs = rhs

    def is_estimable(self):
        is_estimable(self.lhs)
        if self.rhs:
            is_estimable(self.rhs)
        return True

    def is_hint(self):
        return is_hint(self.lhs) or is_hint(self.rhs)

    def __repr__(self):
        if not self.rhs:
            return f'{repr(self.op)}{self.lhs}'
        return f'({self.lhs} {repr(self.op)} {self.rhs})'


class Shape:
    def __init__(self, dims, name='x'):
        super().__init__()
        self.name = name
        self.dims = dims

    def product(self):
        value = 1
        for dim in self.dims or []:
            value = Expr(Operator.MUL, value, dim)
        return value

    def __len__(self):
        return len(self.dims)

    def __repr__(self):
        dims = ', '.join(str(d) if d else '*' for d in self.dims)
        return f' = {dims}'



import abc
import os
import types

import appdirs

from .. import info


class Vars(metaclass=abc.ABCMeta):
    def attach_parser(self, parser):
        for k, v in self._dict().items():
            if isinstance(v, (list, set)):
                parser.add_argument(f'--{k}', nargs='+',
                                    help=getattr(self, f'_{k}_help'))
            else:
                parser.add_argument(f'--{k}', type=type(v),
                                    help=getattr(self, f'_{k}_help'))

    def apply(self, args):
        for k in self._keys():
            v = getattr(args, k)
            if v is None:
                continue
            if isinstance(getattr(self, k), set):
                v = set(v)
            setattr(self, k, v)

    def _keys(self):
        return [k for k in dir(self) if
                not k.startswith('_') and
                not isinstance(getattr(self, k), types.MethodType)]

    def _dict(self):
        return {k: getattr(self, k) for k in self._keys()}

    def __repr__(self):
        return repr(self._dict())


class EnvVars(Vars):
    root = appdirs.user_data_dir(info.APP_NAME, info.APP_AUTHOR)
    _root_help = 'hehehehe'  # TODO: to be implemented

    devices = set()
    _devices_help = 'hehehehe'  # TODO: to be implemented

    def __init__(self):
        super().__init__()
        self._load_env()

    def _load_env(self):
        for key_origin in self._keys():
            key = f'N3_{key_origin.upper()}'
            if key in os.environ:
                value = os.environ[key]
                origin = getattr(self, key_origin)

                if isinstance(origin, (list, set)):
                    value = [v.strip() for v in value.split(',')]
                    if isinstance(origin, set):
                        value = set(value)
                else:
                    raise Exception(
                        f'unexpected type for "{key_origin}": "{type(origin).__name__}"')
                setattr(self, key_origin, value)



import torch.nn as nn

from n3 import ExternNode


class Linear(ExternNode):
    input_channels: int
    output_channels: int

    bias: bool

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._inner = nn.Linear(self.input_channels,
                                self.output_channels,
                                self.bias)

    def forward(self, x):
        return self._inner(x)
